@@ -1,19 +1,19 @@
 /*  -*- c-file-style: "linux" -*-
- * 
- * Copyright (C) 1996-2000 by Andrew Tridgell 
+ *
+ * Copyright (C) 1996-2000 by Andrew Tridgell
  * Copyright (C) Paul Mackerras 1996
  * Copyright (C) 2001, 2002 by Martin Pool <mbp@samba.org>
- * 
+ *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
@@ -22,12 +22,13 @@
 /**
  * @file
  *
- * Utilities used in rsync 
+ * Utilities used in rsync
  **/
 
 #include "rsync.h"
 
 extern int verbose;
+extern struct exclude_list_struct server_exclude_list;
 
 int sanitize_paths = 0;
 
@@ -67,7 +68,7 @@ void set_blocking(int fd)
 /**
  * Create a file descriptor pair - like pipe() but use socketpair if
  * possible (because of blocking issues on pipes).
- * 
+ *
  * Always set non-blocking.
  */
 int fd_pair(int fd[2])
@@ -111,14 +112,14 @@ void print_child_argv(char **cmd)
 
 void out_of_memory(char *str)
 {
-  rprintf(FERROR,"ERROR: out of memory in %s\n",str);
-  exit_cleanup(RERR_MALLOC);
+	rprintf(FERROR, "ERROR: out of memory in %s\n", str);
+	exit_cleanup(RERR_MALLOC);
 }
 
 void overflow(char *str)
 {
-  rprintf(FERROR,"ERROR: buffer overflow in %s\n",str);
-  exit_cleanup(RERR_MALLOC);
+	rprintf(FERROR, "ERROR: buffer overflow in %s\n", str);
+	exit_cleanup(RERR_MALLOC);
 }
 
 
@@ -134,10 +135,10 @@ int set_modtime(char *fname, time_t modtime)
 			fname, (long) modtime,
 			asctime(localtime(&modtime)));
 	}
-	
+
 	{
 #ifdef HAVE_UTIMBUF
-		struct utimbuf tbuf;  
+		struct utimbuf tbuf;
 		tbuf.actime = time(NULL);
 		tbuf.modtime = modtime;
 		return utime(fname,&tbuf);
@@ -167,13 +168,15 @@ int create_directory_path(char *fname, int base_umask)
 {
 	char *p;
 
-	while (*fname == '/') fname++;
-	while (strncmp(fname,"./",2)==0) fname += 2;
+	while (*fname == '/')
+		fname++;
+	while (strncmp(fname, "./", 2) == 0)
+		fname += 2;
 
 	p = fname;
-	while ((p=strchr(p,'/'))) {
+	while ((p = strchr(p,'/')) != NULL) {
 		*p = 0;
-		do_mkdir(fname, 0777 & ~base_umask); 
+		do_mkdir(fname, 0777 & ~base_umask);
 		*p = '/';
 		p++;
 	}
@@ -194,15 +197,13 @@ int create_directory_path(char *fname, int base_umask)
 static int full_write(int desc, char *ptr, size_t len)
 {
 	int total_written;
-	
+
 	total_written = 0;
 	while (len > 0) {
-		int written = write (desc, ptr, len);
+		int written = write(desc, ptr, len);
 		if (written < 0)  {
-#ifdef EINTR
 			if (errno == EINTR)
 				continue;
-#endif
 			return written;
 		}
 		total_written += written;
@@ -227,18 +228,14 @@ static int full_write(int desc, char *ptr, size_t len)
 static int safe_read(int desc, char *ptr, size_t len)
 {
 	int n_chars;
- 
+
 	if (len == 0)
 		return len;
- 
-#ifdef EINTR
+
 	do {
 		n_chars = read(desc, ptr, len);
 	} while (n_chars < 0 && errno == EINTR);
-#else
-	n_chars = read(desc, ptr, len);
-#endif
- 
+
 	return n_chars;
 }
 
@@ -274,7 +271,7 @@ int copy_file(char *source, char *dest, mode_t mode)
 		return -1;
 	}
 
-	while ((len = safe_read(ifd, buf, sizeof(buf))) > 0) {
+	while ((len = safe_read(ifd, buf, sizeof buf)) > 0) {
 		if (full_write(ofd, buf, len) < 0) {
 			rprintf(FERROR,"write %s: %s\n",
 				dest,strerror(errno));
@@ -320,17 +317,15 @@ int robust_unlink(char *fname)
 	char path[MAXPATHLEN];
 
 	rc = do_unlink(fname);
-	if ((rc == 0) || (errno != ETXTBSY))
+	if (rc == 0 || errno != ETXTBSY)
 		return rc;
 
-	strlcpy(path, fname, MAXPATHLEN);
+	if ((pos = strlcpy(path, fname, MAXPATHLEN)) >= MAXPATHLEN)
+		pos = MAXPATHLEN - 1;
 
-	pos = strlen(path);
-	while((path[--pos] != '/') && (pos >= 0))
-		;
-	++pos;
-	strlcpy(&path[pos], ".rsync", MAXPATHLEN-pos);
-	pos += sizeof(".rsync")-1;
+	while (pos > 0 && path[pos-1] != '/')
+		pos--;
+	pos += strlcpy(path+pos, ".rsync", MAXPATHLEN-pos);
 
 	if (pos > (MAXPATHLEN-MAX_RENAMES_DIGITS-1)) {
 		errno = ETXTBSY;
@@ -343,11 +338,12 @@ int robust_unlink(char *fname)
 		sprintf(&path[pos], "%03d", counter);
 		if (++counter >= MAX_RENAMES)
 			counter = 1;
-	} while (((rc = access(path, 0)) == 0) && (counter != start));
+	} while ((rc = access(path, 0)) == 0 && counter != start);
 
-	if (verbose > 0)
+	if (verbose > 0) {
 		rprintf(FINFO,"renaming %s to %s because of text busy\n",
-					    fname, path);
+			fname, path);
+	}
 
 	/* maybe we should return rename()'s exit status? Nah. */
 	if (do_rename(fname, path) != 0) {
@@ -358,18 +354,33 @@ int robust_unlink(char *fname)
 #endif
 }
 
-int robust_rename(char *from, char *to)
+/* Returns 0 on success, -1 on most errors, and -2 if we got an error
+ * trying to copy the file across file systems. */
+int robust_rename(char *from, char *to, int mode)
 {
-#ifndef ETXTBSY
-	return do_rename(from, to);
-#else
-	int rc = do_rename(from, to);
-	if ((rc == 0) || (errno != ETXTBSY))
-		return rc;
-	if (robust_unlink(to) != 0)
-		return -1;
-	return do_rename(from, to);
+	int tries = 4;
+
+	while (tries--) {
+		if (do_rename(from, to) == 0)
+			return 0;
+
+		switch (errno) {
+#ifdef ETXTBSY
+		case ETXTBSY:
+			if (robust_unlink(to) != 0)
+				return -1;
+			break;
 #endif
+		case EXDEV:
+			if (copy_file(from, to, mode) != 0)
+				return -2;
+			do_unlink(from);
+			return 0;
+		default:
+			return -1;
+		}
+	}
+	return -1;
 }
 
 
@@ -380,7 +391,7 @@ static int num_pids;
 pid_t do_fork(void)
 {
 	pid_t newpid = fork();
-	
+
 	if (newpid != 0  &&  newpid != -1) {
 		all_pids[num_pids++] = newpid;
 	}
@@ -454,18 +465,34 @@ int lock_range(int fd, int offset, int len)
 	lock.l_start = offset;
 	lock.l_len = len;
 	lock.l_pid = 0;
-	
+
 	return fcntl(fd,F_SETLK,&lock) == 0;
 }
 
+static int exclude_server_path(char *arg)
+{
+	char *s;
+
+	if (server_exclude_list.head) {
+		for (s = arg; (s = strchr(s, '/')) != NULL; ) {
+			*s = '\0';
+			if (check_exclude(&server_exclude_list, arg, 1) < 0) {
+				/* We must leave arg truncated! */
+				return 1;
+			}
+			*s++ = '/';
+		}
+	}
+	return 0;
+}
 
 static void glob_expand_one(char *s, char **argv, int *argc, int maxargs)
 {
 #if !(defined(HAVE_GLOB) && defined(HAVE_GLOB_H))
 	if (!*s) s = ".";
-	argv[*argc] = strdup(s);
+	s = argv[*argc] = strdup(s);
+	exclude_server_path(s);
 	(*argc)++;
-	return;
 #else
 	extern int sanitize_paths;
 	glob_t globbuf;
@@ -473,54 +500,60 @@ static void glob_expand_one(char *s, char **argv, int *argc, int maxargs)
 
 	if (!*s) s = ".";
 
-	argv[*argc] = strdup(s);
+	s = argv[*argc] = strdup(s);
 	if (sanitize_paths) {
-		sanitize_path(argv[*argc], NULL);
+		sanitize_path(s, NULL);
 	}
 
-	memset(&globbuf, 0, sizeof(globbuf));
-	glob(argv[*argc], 0, NULL, &globbuf);
+	memset(&globbuf, 0, sizeof globbuf);
+	if (!exclude_server_path(s))
+		glob(s, 0, NULL, &globbuf);
 	if (globbuf.gl_pathc == 0) {
 		(*argc)++;
 		globfree(&globbuf);
 		return;
 	}
-	for (i=0; i<(maxargs - (*argc)) && i < (int) globbuf.gl_pathc;i++) {
-		if (i == 0) free(argv[*argc]);
-		argv[(*argc) + i] = strdup(globbuf.gl_pathv[i]);
-		if (!argv[(*argc) + i]) out_of_memory("glob_expand");
+	for (i = 0; i < maxargs - *argc && i < (int)globbuf.gl_pathc; i++) {
+		if (i == 0)
+			free(s);
+		argv[*argc + i] = strdup(globbuf.gl_pathv[i]);
+		if (!argv[*argc + i])
+			out_of_memory("glob_expand");
 	}
 	globfree(&globbuf);
-	(*argc) += i;
+	*argc += i;
 #endif
 }
 
+/* This routine is only used in daemon mode. */
 void glob_expand(char *base1, char **argv, int *argc, int maxargs)
 {
 	char *s = argv[*argc];
 	char *p, *q;
 	char *base = base1;
+	int base_len = strlen(base);
 
 	if (!s || !*s) return;
 
-	if (strncmp(s, base, strlen(base)) == 0) {
-		s += strlen(base);
-	}
+	if (strncmp(s, base, base_len) == 0)
+		s += base_len;
 
 	s = strdup(s);
 	if (!s) out_of_memory("glob_expand");
 
 	if (asprintf(&base," %s/", base1) <= 0) out_of_memory("glob_expand");
+	base_len++;
 
 	q = s;
-	while ((p = strstr(q,base)) && ((*argc) < maxargs)) {
+	while ((p = strstr(q,base)) != NULL && *argc < maxargs) {
 		/* split it at this point */
 		*p = 0;
 		glob_expand_one(q, argv, argc, maxargs);
-		q = p+strlen(base);
+		q = p + base_len;
 	}
 
-	if (*q && (*argc < maxargs)) glob_expand_one(q, argv, argc, maxargs);
+	if (*q && *argc < maxargs)
+		glob_expand_one(q, argv, argc, maxargs);
 
 	free(s);
 	free(base);
@@ -538,6 +571,58 @@ void strlower(char *s)
 	}
 }
 
+/* Join strings p1 & p2 into "dest" with a guaranteed '/' between them.  (If
+ * p1 ends with a '/', no extra '/' is inserted.)  Returns the length of both
+ * strings + 1 (if '/' was inserted), regardless of whether the null-terminated
+ * string fits into destsize. */
+size_t pathjoin(char *dest, size_t destsize, const char *p1, const char *p2)
+{
+	size_t len = strlcpy(dest, p1, destsize);
+	if (len < destsize - 1) {
+		if (!len || dest[len-1] != '/')
+			dest[len++] = '/';
+		if (len < destsize - 1)
+			len += strlcpy(dest + len, p2, destsize - len);
+		else {
+			dest[len] = '\0';
+			len += strlen(p2);
+		}
+	}
+	else
+		len += strlen(p2) + 1; /* Assume we'd insert a '/'. */
+	return len;
+}
+
+/* Join any number of strings together, putting them in "dest".  The return
+ * value is the length of all the strings, regardless of whether the null-
+ * terminated whole fits in destsize.  Your list of string pointers must end
+ * with a NULL to indicate the end of the list. */
+size_t stringjoin(char *dest, size_t destsize, ...)
+{
+	va_list ap;
+	size_t len, ret = 0;
+	const char *src;
+
+	va_start(ap, destsize);
+	while (1) {
+		if (!(src = va_arg(ap, const char *)))
+			break;
+		len = strlen(src);
+		ret += len;
+		if (destsize > 1) {
+			if (len >= destsize)
+				len = destsize - 1;
+			memcpy(dest, src, len);
+			destsize -= len;
+			dest += len;
+		}
+	}
+	*dest = '\0';
+	va_end(ap);
+
+	return ret;
+}
+
 void clean_fname(char *name)
 {
 	char *p;
@@ -549,7 +634,7 @@ void clean_fname(char *name)
 	while (modified) {
 		modified = 0;
 
-		if ((p=strstr(name,"/./"))) {
+		if ((p = strstr(name,"/./")) != NULL) {
 			modified = 1;
 			while (*p) {
 				p[0] = p[2];
@@ -557,7 +642,7 @@ void clean_fname(char *name)
 			}
 		}
 
-		if ((p=strstr(name,"//"))) {
+		if ((p = strstr(name,"//")) != NULL) {
 			modified = 1;
 			while (*p) {
 				p[0] = p[1];
@@ -565,14 +650,14 @@ void clean_fname(char *name)
 			}
 		}
 
-		if (strncmp(p=name,"./",2) == 0) {      
+		if (strncmp(p = name, "./", 2) == 0) {
 			modified = 1;
 			do {
 				p[0] = p[2];
 			} while (*p++);
 		}
 
-		l = strlen(p=name);
+		l = strlen(p = name);
 		if (l > 1 && p[l-1] == '/') {
 			modified = 1;
 			p[l-1] = 0;
@@ -631,7 +716,7 @@ void sanitize_path(char *p, char *reldir)
 		 * both p (and sanp if the original had a slash) should
 		 * always be left pointing after a slash
 		 */
-		if ((*p == '.') && ((*(p+1) == '/') || (*(p+1) == '\0'))) {
+		if (*p == '.' && (p[1] == '/' || p[1] == '\0')) {
 			/* skip "." component */
 			while (*++p == '/') {
 				/* skip following slashes */
@@ -640,10 +725,9 @@ void sanitize_path(char *p, char *reldir)
 			continue;
 		}
 		allowdotdot = 0;
-		if ((*p == '.') && (*(p+1) == '.') &&
-			    ((*(p+2) == '/') || (*(p+2) == '\0'))) {
+		if (*p == '.' && p[1] == '.' && (p[2] == '/' || p[2] == '\0')) {
 			/* ".." component followed by slash or end */
-			if ((depth > 0) && (sanp == start)) {
+			if (depth > 0 && sanp == start) {
 				/* allow depth levels of .. at the beginning */
 				--depth;
 				allowdotdot = 1;
@@ -654,7 +738,7 @@ void sanitize_path(char *p, char *reldir)
 				if (sanp != start) {
 					/* back up sanp one level */
 					--sanp; /* now pointing at slash */
-					while ((sanp > start) && (*(sanp - 1) != '/')) {
+					while (sanp > start && sanp[-1] != '/') {
 						/* skip back up to slash */
 						sanp--;
 					}
@@ -665,7 +749,7 @@ void sanitize_path(char *p, char *reldir)
 		while (1) {
 			/* copy one component through next slash */
 			*sanp++ = *p++;
-			if ((*p == '\0') || (*(p-1) == '/')) {
+			if (*p == '\0' || p[-1] == '/') {
 				while (*p == '/') {
 					/* skip multiple slashes */
 					p++;
@@ -678,7 +762,7 @@ void sanitize_path(char *p, char *reldir)
 			start = sanp;
 		}
 	}
-	if ((sanp == start) && !allowdotdot) {
+	if (sanp == start && !allowdotdot) {
 		/* ended up with nothing, so put in "." component */
 		/*
 		 * note that the !allowdotdot doesn't prevent this from
@@ -691,59 +775,146 @@ void sanitize_path(char *p, char *reldir)
 	*sanp = '\0';
 }
 
+/* Works much like sanitize_path(), with these differences:  (1) a new buffer
+ * is allocated for the sanitized path rather than modifying it in-place; (2)
+ * a leading slash gets transformed into the rootdir value (which can be empty
+ * or NULL if you just want the slash to get dropped); (3) no "reldir" can be
+ * specified. */
+char *alloc_sanitize_path(const char *path, const char *rootdir)
+{
+	char *buf;
+	int rlen, plen = strlen(path);
 
-static char curr_dir[MAXPATHLEN];
+	if (*path == '/' && rootdir) {
+		rlen = strlen(rootdir);
+		if (rlen == 1)
+			path++;
+	} else
+		rlen = 0;
+	if (!(buf = new_array(char, rlen + plen + 1)))
+		out_of_memory("alloc_sanitize_path");
+	if (rlen)
+		memcpy(buf, rootdir, rlen);
+	memcpy(buf + rlen, path, plen + 1);
+
+	if (rlen > 1)
+		rlen++;
+	sanitize_path(buf + rlen, NULL);
+	if (rlen && buf[rlen] == '.' && buf[rlen+1] == '\0') {
+		if (rlen > 1)
+			rlen--;
+		buf[rlen] = '\0';
+	}
+
+	return buf;
+}
+
+char curr_dir[MAXPATHLEN];
+unsigned int curr_dir_len;
 
 /**
- * Like chdir() but can be reversed with pop_dir() if @p save is set.
- * It is also much faster as it remembers where we have been.
+ * Like chdir(), but it keeps track of the current directory (in the
+ * global "curr_dir"), and ensures that the path size doesn't overflow.
+ * Also cleans the path using the clean_fname() function.
  **/
-char *push_dir(char *dir, int save)
+int push_dir(char *dir)
 {
-	char *ret = curr_dir;
 	static int initialised;
+	unsigned int len;
 
 	if (!initialised) {
 		initialised = 1;
-		getcwd(curr_dir, sizeof(curr_dir)-1);
+		getcwd(curr_dir, sizeof curr_dir - 1);
+		curr_dir_len = strlen(curr_dir);
 	}
 
-	if (!dir) return NULL; /* this call was probably just to initialize */
+	if (!dir)	/* this call was probably just to initialize */
+		return 0;
 
-	if (chdir(dir)) return NULL;
+	len = strlen(dir);
+	if (len == 1 && *dir == '.')
+		return 1;
 
-	if (save) {
-		ret = strdup(curr_dir);
-	}
+	if ((*dir == '/' ? len : curr_dir_len + 1 + len) >= sizeof curr_dir)
+		return 0;
+
+	if (chdir(dir))
+		return 0;
 
 	if (*dir == '/') {
-		strlcpy(curr_dir, dir, sizeof(curr_dir));
+		memcpy(curr_dir, dir, len + 1);
+		curr_dir_len = len;
 	} else {
-		strlcat(curr_dir,"/", sizeof(curr_dir));
-		strlcat(curr_dir,dir, sizeof(curr_dir));
+		curr_dir[curr_dir_len++] = '/';
+		memcpy(curr_dir + curr_dir_len, dir, len + 1);
+		curr_dir_len += len;
 	}
 
 	clean_fname(curr_dir);
 
-	return ret;
+	return 1;
 }
 
-/** Reverse a push_dir() call */
+/**
+ * Reverse a push_dir() call.  You must pass in an absolute path
+ * that was copied from a prior value of "curr_dir".
+ **/
 int pop_dir(char *dir)
 {
-	int ret;
+	if (chdir(dir))
+		return 0;
 
-	ret = chdir(dir);
-	if (ret) {
-		free(dir);
-		return ret;
+	curr_dir_len = strlcpy(curr_dir, dir, sizeof curr_dir);
+	if (curr_dir_len >= sizeof curr_dir)
+		curr_dir_len = sizeof curr_dir - 1;
+
+	return 1;
+}
+
+/**
+ * Return a quoted string with the full pathname of the indicated filename.
+ * The string " (in MODNAME)" may also be appended.  The returned pointer
+ * remains valid until the next time full_fname() is called.
+ **/
+char *full_fname(char *fn)
+{
+	extern int module_id;
+	static char *result = NULL;
+	char *m1, *m2, *m3;
+	char *p1, *p2;
+
+	if (result)
+		free(result);
+
+	if (*fn == '/')
+		p1 = p2 = "";
+	else {
+		p1 = curr_dir;
+		p2 = "/";
 	}
+	if (module_id >= 0) {
+		m1 = " (in ";
+		m2 = lp_name(module_id);
+		m3 = ")";
+		if (*p1) {
+			if (!lp_use_chroot(module_id)) {
+				char *p = lp_path(module_id);
+				if (*p != '/' || p[1])
+					p1 += strlen(p);
+			}
+			if (!*p1)
+				p2++;
+			else
+				p1++;
+		}
+		else
+			fn++;
+	} else
+		m1 = m2 = m3 = "";
 
-	strlcpy(curr_dir, dir, sizeof(curr_dir));
+	asprintf(&result, "\"%s%s%s\"%s%s%s", p1, p2, fn, m1, m2, m3);
 
-	free(dir);
-
-	return 0;
+	return result;
 }
 
 /** We need to supply our own strcmp function for file list comparisons
@@ -756,7 +927,7 @@ int u_strcmp(const char *cs1, const char *cs2)
 	while (*s1 && *s2 && (*s1 == *s2)) {
 		s1++; s2++;
 	}
-	
+
 	return (int)*s1 - (int)*s2;
 }
 
@@ -797,7 +968,7 @@ int unsafe_symlink(const char *dest, const char *src)
 	/* find out what our safety margin is */
 	for (name = src; (slash = strchr(name, '/')) != 0; name = slash+1) {
 		if (strncmp(name, "../", 3) == 0) {
-			depth=0;
+			depth = 0;
 		} else if (strncmp(name, "./", 2) == 0) {
 			/* nothing */
 		} else {
@@ -835,9 +1006,9 @@ char *timestring(time_t t)
 	struct tm *tm = localtime(&t);
 
 #ifdef HAVE_STRFTIME
-	strftime(TimeBuf,sizeof(TimeBuf)-1,"%Y/%m/%d %T",tm);
+	strftime(TimeBuf, sizeof TimeBuf - 1, "%Y/%m/%d %H:%M:%S", tm);
 #else
-	strlcpy(TimeBuf, asctime(tm), sizeof(TimeBuf));
+	strlcpy(TimeBuf, asctime(tm), sizeof TimeBuf);
 #endif
 
 	if (TimeBuf[strlen(TimeBuf)-1] == '\n') {
@@ -856,21 +1027,21 @@ char *timestring(time_t t)
  **/
 int msleep(int t)
 {
-	int tdiff=0;
-	struct timeval tval,t1,t2;  
+	int tdiff = 0;
+	struct timeval tval, t1, t2;
 
 	gettimeofday(&t1, NULL);
 	gettimeofday(&t2, NULL);
-  
+
 	while (tdiff < t) {
 		tval.tv_sec = (t-tdiff)/1000;
 		tval.tv_usec = 1000*((t-tdiff)%1000);
- 
+
 		errno = 0;
 		select(0,NULL,NULL, NULL, &tval);
 
 		gettimeofday(&t2, NULL);
-		tdiff = (t2.tv_sec - t1.tv_sec)*1000 + 
+		tdiff = (t2.tv_sec - t1.tv_sec)*1000 +
 			(t2.tv_usec - t1.tv_usec)/1000;
 	}
 
@@ -916,7 +1087,7 @@ int _Insure_trap_error(int a1, int a2, int a3, int a4, int a5, int a6)
 	int ret;
 	char *cmd;
 
-	asprintf(&cmd, "/usr/X11R6/bin/xterm -display :0 -T Panic -n Panic -e /bin/sh -c 'cat /tmp/ierrs.*.%d ; gdb /proc/%d/exe %d'", 
+	asprintf(&cmd, "/usr/X11R6/bin/xterm -display :0 -T Panic -n Panic -e /bin/sh -c 'cat /tmp/ierrs.*.%d ; gdb /proc/%d/exe %d'",
 		getpid(), getpid(), getpid());
 
 	if (!fn) {
